from enum import StrEnum
import json
import paho.mqtt.client as mqtt
import platform

class HassioDevice:
    identifiers: list[str] = None
    name: str = None
    manufacturer: str = None
    model: str = None
    serial_number: str = None
    sw_version: str = None
    configuration_url: str = None

    def __init__(self, identifiers: list[str], name: str, manufacturer: str, model: str, serial_number: str, sw_version: str, configuration_url: str):
        self.identifiers = identifiers
        self.name = name
        self.manufacturer = manufacturer
        self.model = model
        self.serial_number = serial_number
        self.sw_version = sw_version
        self.configuration_url = configuration_url

    @property
    def discovery_data(self) -> dict:
        return {
            "identifiers": self.identifiers,
            "name": self.name,
            "manufacturer": self.manufacturer,
            "model": self.model,
            "serial_number": self.serial_number,
            "sw_version": self.sw_version,
            "configuration_url": self.configuration_url
        }

class HassioComponent(StrEnum):
    SENSOR = 'sensor'
    BINARY_SENSOR = 'binary_sensor'

class HassioEntity:
    component: HassioComponent = None
    name: str = None
    device_class: str = None
    unique_id: str = None
    device: HassioDevice = None

    def __init__(self, component: HassioComponent, name: str, device_class: str | None, unique_id: str, device: HassioDevice):
        self.component  = component
        self.name = name
        self.device_class = device_class
        self.unique_id = unique_id
        self.device = device

    @property
    def topic(self) -> str:
        return f"{self.component}/{self.unique_id}"

    @property
    def discovery_data(self) -> dict:
        return {
            "name": self.name,
            "device_class": self.device_class,
            "unique_id": self.unique_id
        }

class HassioBinarySensorEntity(HassioEntity):
    def __init__(self, name: str, device_class: str | None, unique_id: str, device: HassioDevice):
        super().__init__(HassioComponent.BINARY_SENSOR, name, device_class, unique_id, device)

class HassioSensorEntity(HassioEntity):
    unit_of_measurement: str = None
    state_class: str = None

    def __init__(self, name: str, device_class: str | None, unit_of_measurement: str | None, state_class: str | None, unique_id: str, device: HassioDevice):
        super().__init__(HassioComponent.SENSOR, name, device_class, unique_id, device)
        self.unit_of_measurement = unit_of_measurement
        self.state_class = state_class

    @property
    def discovery_data(self) -> dict:
        return {
            **super().discovery_data,
            "unit_of_measurement": self.unit_of_measurement,
            "state_class": self.state_class
        }

class HassioClient:
    def __init__(self, host: str = 'homeassistant.local', port: int = 1883, discovery_prefix: str = 'homeassistant', username: str = None, password: str = None, kv_opts = {}):
        self._host = host
        self._port = port
        self._prefix = discovery_prefix
        self._mqtt = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2, platform.node(), **kv_opts)
        self._mqtt.username_pw_set(username, password)

    def _entity_topic(self, entity: HassioEntity) -> str:
        return f"{self._prefix}/{entity.topic}"

    def _discovery_topic(self, entity: HassioEntity) -> str:
        return f"{self._entity_topic(entity)}/config"

    def _state_topic(self, entity: HassioEntity) -> str:
        return f"{self._entity_topic(entity)}/state"
    
    def _availability_topic(self, entity: HassioEntity) -> str:
        return f"{self._entity_topic(entity)}/availability"

    def start(self) -> None:
        res = self._mqtt.connect(self._host, self._port)
        if res != mqtt.MQTTErrorCode.MQTT_ERR_SUCCESS:
            raise ConnectionError(res)
        self._mqtt.loop_start()

    def stop(self) -> None:
        self._mqtt.loop_stop()

    def register(self, entity: HassioEntity, device_ident_only: bool = False) -> None:
        payload = json.dumps({
            **entity.discovery_data,
            "state_topic": self._state_topic(entity),
            "availability_topic": self._availability_topic(entity),
            "device": {
                "identifiers": entity.device.identifiers
            } if device_ident_only else {
                **entity.device.discovery_data
            }
        })
        self._mqtt.publish(self._discovery_topic(entity), payload, qos=1, retain=True).wait_for_publish()

    def unregister(self, entity: HassioEntity) -> None:
        self._mqtt.publish(self._discovery_topic(entity), '', qos=1).wait_for_publish()

    def update_availability(self, entity: HassioEntity, available: bool) -> None:
        self._mqtt.publish(self._availability_topic(entity), 'online' if available else 'offline').wait_for_publish()

    def update_state(self, entity: HassioEntity, payload) -> None:
        self._mqtt.publish(self._state_topic(entity), payload).wait_for_publish()


from euc2mqtt.HassioAPI import HassioComponent
from .HassioAPI import HassioDevice, HassioSensorEntity, HassioBinarySensorEntity, HassioEntity
from .Utility import id_format

class UpsDevice(HassioDevice):
    def __init__(self, name: str, data: dict, config_url: str):
        super().__init__(identifiers=[data['status']['serialNumber']],
                         name=name if name is not None else data['deviceInfo']['name'],
                         manufacturer=data['sysInfo']['manufacturer'],
                         model=f"{data['status']['product']} {data['status']['model']}",
                         serial_number=data['status']['serialNumber'],
                         sw_version=data['status']['firmwareVersion'],
                         configuration_url=config_url)

class UpsEntityMixin:
    def __init__(self, path: str) -> None:
        self._path = path

    def update(self, data: dict):
        for p in self._path.split('.'):
            if p in data:
                data = data[p]
            else:
                return None
        return self.on_update(data)

    def on_update(self, data):
        return None

class UpsEntity(HassioEntity, UpsEntityMixin):
    def __init__(self, path: str, component: HassioComponent, name: str, device_class: str | None, unique_id: str, device: HassioDevice):
        HassioEntity.__init__(self, component, name, device_class, unique_id, device)
        UpsEntityMixin.__init__(self, path)

class UpsSensorEntity(HassioSensorEntity, UpsEntityMixin):
    def __init__(self, path: str, name: str, device_class: str | None, unit_of_measurement: str | None, state_class: str | None, unique_id: str, device: HassioDevice, update_func):
        HassioSensorEntity.__init__(self, name, device_class, unit_of_measurement, state_class, unique_id, device)
        UpsEntityMixin.__init__(self, path)
        self._update_func = update_func

    def on_update(self, data):
        return self._update_func(data)

class UpsBinarySensorEntity(HassioBinarySensorEntity, UpsEntityMixin):
    def __init__(self, path: str, name: str, device_class: str | None, unique_id: str, device: HassioDevice):
        HassioBinarySensorEntity.__init__(self, name, device_class, unique_id, device)
        UpsEntityMixin.__init__(self, path)

    def on_update(self, data):
        return 'ON' if bool(data) else 'OFF'

class UpsEntityCollection:
    def __init__(self, name: str, device: UpsDevice):
        self._name = name
        self._device = device
        self._entities = []

    @property
    def entities(self) -> list[UpsEntity]:
        return self._entities

    def add_sensor(self, path: str, name: str, device_class: str | None, unit_of_measurement: str | None = None, state_class: str | None = 'measurement', update_func = lambda x: x):
        sensor = UpsSensorEntity(path=path,
                                   name=name,
                                   device_class=device_class,
                                   unit_of_measurement=unit_of_measurement,
                                   state_class=state_class,
                                   unique_id=id_format(f"{self._name} {name}").lower(),
                                   device=self._device,
                                   update_func=update_func)
        self._entities.append(sensor)

    def add_binary_sensor(self, path: str, name: str, device_class: str | None):
        binary_sensor = UpsBinarySensorEntity(path=path,
                                                name=name,
                                                device_class=device_class,
                                                unique_id=id_format(f"{self._name} {name}").lower(),
                                                device=self._device)
        self._entities.append(binary_sensor)


from .EatonAPI import EatonClient
from .HassioAPI import HassioClient
from .InteropE2H import UpsDevice, UpsEntity, UpsEntityCollection

class UpdateHandler:
    on_register = []
    on_com_lost = []
    on_update = []
    on_unregister = []

    def __init__(self, name: str, eaton_host: str = 'localhost', eaton_port: int = 4679, mqtt_host: str = 'homeassistant.local', mqtt_port: int = 1883, mqtt_username: str = None, mqtt_password: str = None, mqtt_opts = {}):
        self._eaton_client = EatonClient(eaton_host, eaton_port)
        self._hassio_client = HassioClient(mqtt_host, mqtt_port, username=mqtt_username, password=mqtt_password, kv_opts=mqtt_opts)
        self._name = name
        self._last_update = 0
        self._collection = None

    def _add_sensors(self):
        c = self._collection
        c.add_sensor('status.outputPower', 'Output Power', 'power', 'W')
        c.add_sensor('status.energy', 'Energy', 'energy', 'kWh', 'total', lambda x: round(x / 3600.0 / 1000.0, 2))
        c.add_sensor('status.batteryCapacity', 'Battery Capacity', 'battery', '%')
        c.add_sensor('status.batteryRunTime', 'Battery Run Time', 'duration', 's')
        c.add_sensor('status.remainingCapacityLimit', 'Remaining Capacity Limit', None, '%')
        c.add_sensor('status.runTimeToShutdown', 'Run Time to Shutdown', 'duration', 's', update_func=lambda x: x if x > 0 else None)
        c.add_binary_sensor('status.acPresent', 'AC Present', 'power')
        c.add_binary_sensor('status.charging', 'Charging', 'battery_charging')
        c.add_binary_sensor('status.overload', 'Overload', 'problem')
        c.add_binary_sensor('status.outputStatus', 'Output', 'power')
        c.add_binary_sensor('status.internalFailure', 'Internal Failure', 'problem')
        c.add_binary_sensor('status.batteryLow', 'Battery Low', 'battery')
        c.add_binary_sensor('status.batteryFault', 'Battery Fault', 'problem')
        c.add_binary_sensor('status.shutdownImminent', 'Shutdown Imminent', 'problem')

    def _notify_register(self, entity: UpsEntity):
        for fn in self.on_register:
            fn(entity)

    def _notify_com_lost(self):
        for fn in self.on_com_lost:
            fn()

    def _notify_update(self, entity: UpsEntity, value):
        for fn in self.on_update:
            fn(entity, value)

    def _notify_unregister(self, entity: UpsEntity):
        for fn in self.on_register:
            fn(entity)

    def _register_entities(self):
        for i in range(0, len(self._collection.entities)):
            e = self._collection.entities[i]
            self._hassio_client.register(e, i > 0)
            self._notify_register(e)

    def _unregister_entities(self):
        for e in self._collection.entities:
            self._hassio_client.unregister(e)
            self._notify_unregister(e)

    def start(self):
        # Get initial data, create device and sensors
        device = UpsDevice(self._name,
                           self._eaton_client.get_data(),
                           self._eaton_client.admin_url)
        if self._name is None:
            self._name = device.name
        self._collection = UpsEntityCollection(self._name, device)
        self._add_sensors()
        # Start MQTT and register entities with Hassio
        self._hassio_client.start()
        self._register_entities()
        # Prepare update loop
        self._last_update = 0

    def update(self, full_update: bool = False):
        data = self._eaton_client.get_data(0 if full_update else self._last_update)
        if 'lastUpdate' in data:
            # Update reference timestamp
            self._last_update = data['lastUpdate']
        if ('status' in data) and ('comLost' in data['status']) and bool(data['status']['comLost']):
            # Communication loss
            for e in self._collection.entities:
                self._hassio_client.update_availability(e, False)
            self._notify_com_lost()
        for e in self._collection.entities:
            # Update sensors
            result = e.update(data)
            if result is not None:
                self._hassio_client.update_availability(e, True)
                self._hassio_client.update_state(e, result)
                self._notify_update(e, result)

    def stop(self):
        self._unregister_entities()
        self._hassio_client.stop()


import argparse
import logging
import sys
import time
from . import UpdateHandler

class DefaultConfig:
    EUC_HOST: str = 'localhost'
    EUC_PORT: int = 4679
    HASSIO_HOST: str = 'homeassistant.local'
    HASSIO_PORT: int = 1883
    INTERVAL: int = 30
    FULL_UPDATE_INTERVAL: int = 9

def main():
    parser = argparse.ArgumentParser(prog='euc2mqtt', description='MQTT Publisher for Eaton UPS Companion status messages to Home Assistant. See https://github.com/islandcontroller/euc2mqtt for more info!')
    parser.add_argument('--name', help='Device name', type=str, default=None)
    parser.add_argument('--mqtt', help='MQTT broker hostname and port (hostname[:port])', type=str, default=f"{DefaultConfig.HASSIO_HOST}:{DefaultConfig.HASSIO_PORT}")
    parser.add_argument('--euc', help='Eaton UPS Companion hostname and port (hostname[:port])', type=str, default=f"{DefaultConfig.EUC_HOST}:{DefaultConfig.EUC_PORT}")
    parser.add_argument('--username', help='Username for MQTT broker authentication', type=str, default=None)
    parser.add_argument('--password', help='Password for MQTT broker authentication', type=str, default=None)
    parser.add_argument('--interval', help='Update interval in seconds', type=int, default=DefaultConfig.INTERVAL)
    parser.add_argument('--full-update', help='Number of incremental dataset fetches between full updates', type=int, default=DefaultConfig.FULL_UPDATE_INTERVAL)
    parser.add_argument('--logfile', help='Output log messages to a file', type=str, default=None)
    parser.add_argument('--verbose', help='Enable verbose logging', action='store_true')
    args = parser.parse_args()

    logging.basicConfig(filename=args.logfile,
                        level=logging.DEBUG if args.verbose else logging.INFO,
                        format='%(asctime)s %(levelname)-8s %(message)s',
                        datefmt='%Y-%m-%dT%H:%M:%S%z')
    sys.excepthook = lambda type, value, tb: logging.log(logging.ERROR, f"Caught unhandled exception: {value}")
    
    eaton_conf = args.euc.split(':')
    eaton_host = eaton_conf[0]
    eaton_port = DefaultConfig.EUC_PORT if len(eaton_conf) < 2 else int(eaton_conf[1], base=10)
    logging.log(logging.DEBUG, f"Using EUC service at {eaton_host}:{eaton_port}")

    mqtt_conf = args.mqtt.split(':')
    mqtt_host = mqtt_conf[0]
    mqtt_port = DefaultConfig.HASSIO_PORT if len(mqtt_conf) < 2 else int(mqtt_conf[1], base=10)
    logging.log(logging.DEBUG, f"Using MQTT broker at {mqtt_host}:{mqtt_port}")

    logging.log(logging.INFO, 'Configuring update handler...')
    handler = UpdateHandler(args.name, eaton_host, eaton_port, mqtt_host, mqtt_port, args.username, args.password)
    handler.on_com_lost = [lambda: logging.log(logging.WARNING, 'EUC lost communication with UPS')]
    handler.on_register = [lambda e: logging.log(logging.DEBUG, f"Entity registered: {e.unique_id}")]
    handler.on_update = [lambda e, v: logging.log(logging.DEBUG, f"Entity updated: {e.unique_id} = {v}")]
    handler.on_unregister = [lambda e: logging.log(logging.DEBUG, f"Entity unregistered: {e.unique_id}")]

    logging.log(logging.INFO, f"Starting update loop ({args.interval}s update interval)...")
    handler.start()
    update_count = 0
    while True:
        try:
            logging.log(logging.DEBUG, f"Fetching {'full' if update_count == 0 else 'incremental'} dataset")
            handler.update(update_count == 0)
            if update_count == 0:
                update_count = max(args.full_update, 0)
            else:
                update_count -= 1
            time.sleep(args.interval)
        except KeyboardInterrupt:
            logging.log(logging.WARNING, 'Caught KeyboardInterrupt, exiting...')
            break
    logging.log(logging.INFO, 'Stopping update loop...')
    handler.stop()


# Prior to pip v21.1, a setup.py script was required to be compatible with development mode. With late versions of pip, projects without setup.py may be installed in this mode.

from setuptools import setup

setup()


from enum import StrEnum
import requests

class EatonClient:
    """EatonClient([hostname, [port]])

    API Client for the Eaton UPS Companion webservice.

    This client grants access to API functions of the Eaton UPS Companion
    webservice. The service is installed as part of the Eaton UPS Companion
    software, and is usually accessible at address 'http://localhost:4679/'.
    The API entrypoint is located under path `/euc-data.js`,

    Usage:
        ```python
        client = EatonClient()
        data = client.get_data()
        print(f"AC power status: {data['status']['acPresent']}")
        ```
    """

    API_ENDP_PATH: str = '/euc-data.js'

    class Command(StrEnum):
        GET_DATA = 'getData'
        SET_DATA = 'setData'
        CLEAR_LOGS = 'clearLogs'

    def __init__(self, hostname: str = 'localhost', port: int = 4679):
        """Create a new client.

        Args:
            hostname (str, optional): Hostname of the UPS Companion webservice. Defaults to 'localhost'.
            port (int, optional): Webservice port number. Defaults to 4679.
        """
        self._base_url =  f"http://{hostname}:{port}"
        self._api_endpoint = f"{self._base_url}{self.API_ENDP_PATH}"

    def _transfer(self, cmd: Command, data: dict = {}) -> dict:
        response = requests.post(self._api_endpoint, json={'cmd': cmd, **data})
        response.raise_for_status()
        return response.json()

    @property
    def admin_url(self) -> str:
        """Get the Eaton UPS Companion Webadmin URL."""
        return f"{self._base_url}/"

    def get_data(self, from_date: int = 0) -> dict:
        """Get system and UPS status data.

        Args:
            from_date (int, optional): Timestamp of last update (see Usage below). Defaults to 0.

        Returns:
            dict: Parsed JSON data structure containing system and UPS status data.

        Usage:
            ```python
            client = EatonClient()

            last_update = 0 # Request all data
            while (True):
                data = client.get_data(last_update)
                last_update = data['lastUpdate'] # Only fetch changes on next query
                # ...
            ```
        """
        return self._transfer(self.Command.GET_DATA, {'fromDate': from_date})

    def set_data(self, item_path: str, value: str) -> None:
        """Set configuration item value.

        Args:
            item_path (str): Item path as specified in the EUC web UI, formatted as `<subset>.<item_id>`.
            value (str): New item value.

        Usage:
            ```python
            client = EatonClient()
            client.set_data('systemCfg.systemTray', True)
            ```
        """
        [subset, item_id] = item_path.split('.')
        self._transfer(self.Command.SET_DATA, {'subset': subset, 'itemID': item_id, 'itemValue': value})

    def clear_logs(self) -> None:
        """Clear UPS Companion event history.

        Usage:
            ```python
            client = EatonClient()
            client.clear_logs()
            ```
        """
        self._transfer(self.Command.CLEAR_LOGS)
